@@ -45,14 +45,14 @@ int main(int argc, char **argv) {
   usage += " [text.fst [binary.fst]]\n";
 
   std::set_new_handler(FailedNewHandler);
-  SetFlags(usage.c_str(), &argc, &argv, true);
+  SET_FLAGS(usage.c_str(), &argc, &argv, true);
   if (argc > 3) {
     ShowUsage();
     return 1;
   }
 
   const char *source = "standard input";
-  istream *istrm = &std::cin;
+  istream *istrm = &cin;
   if (argc > 1 && strcmp(argv[1], "-") != 0) {
     source = argv[1];
     istrm = new fst::ifstream(argv[1]);
@@ -85,7 +85,7 @@ int main(int argc, char **argv) {
                 FLAGS_acceptor, FLAGS_keep_isymbols, FLAGS_keep_osymbols,
                 FLAGS_keep_state_numbering, FLAGS_allow_negative_labels);
 
-  if (istrm != &std::cin)
+  if (istrm != &cin)
     delete istrm;
 
   return 0;
