@@ -25,11 +25,13 @@
 #include <string>
 #include <vector>
 using std::vector;
+
 #include <fst/expanded-fst.h>
 #include <fst/fst-decl.h>  // For optional argument declarations
 #include <fst/test-properties.h>
 #include <fst/util.h>
 
+
 namespace fst {
 
 template <class A, class U> class ConstFst;
@@ -46,7 +48,6 @@ class ConstFstImpl : public FstImpl<A> {
   using FstImpl<A>::SetType;
   using FstImpl<A>::SetProperties;
   using FstImpl<A>::Properties;
-  using FstImpl<A>::WriteHeader;
 
   typedef A Arc;
   typedef typename A::Weight Weight;
@@ -105,6 +106,8 @@ class ConstFstImpl : public FstImpl<A> {
   }
 
  private:
+  friend class ConstFst<A, U>;  // Allow finding narcs_, nstates_ during Write
+
   // States implemented by array *states_ below, arcs by (single) *arcs_.
   struct State {
     Weight final;                // Final weight
@@ -117,11 +120,14 @@ class ConstFstImpl : public FstImpl<A> {
 
   // Properties always true of this Fst class
   static const uint64 kStaticProperties = kExpanded;
-  // Current file format version
-  static const int kFileVersion = 1;
+  // Current unaligned file format version. The unaligned version was added and
+  // made the default since the aligned version does not work on pipes.
+  static const int kFileVersion = 2;
+  // Current aligned file format version
+  static const int kAlignedFileVersion = 1;
   // Minimum file format version supported
   static const int kMinFileVersion = 1;
-  // Byte alignment for states and arcs in file format
+  // Byte alignment for states and arcs in file format (version 1 only)
   static const int kFileAlign = 16;
 
   State *states_;                // States represenation
@@ -138,6 +144,8 @@ const uint64 ConstFstImpl<A, U>::kStaticProperties;
 template <class A, class U>
 const int ConstFstImpl<A, U>::kFileVersion;
 template <class A, class U>
+const int ConstFstImpl<A, U>::kAlignedFileVersion;
+template <class A, class U>
 const int ConstFstImpl<A, U>::kMinFileVersion;
 template <class A, class U>
 const int ConstFstImpl<A, U>::kFileAlign;
@@ -152,8 +160,6 @@ ConstFstImpl<A, U>::ConstFstImpl(const Fst<A> &fst) : nstates_(0), narcs_(0) {
     type += size;
   }
   SetType(type);
-  uint64 copy_properties = fst.Properties(kCopyProperties, true);
-  SetProperties(copy_properties | kStaticProperties);
   SetInputSymbols(fst.InputSymbols());
   SetOutputSymbols(fst.OutputSymbols());
   start_ = fst.Start();
@@ -190,68 +196,58 @@ ConstFstImpl<A, U>::ConstFstImpl(const Fst<A> &fst) : nstates_(0), narcs_(0) {
       arcs_[pos++] = arc;
     }
   }
+  SetProperties(fst.Properties(kCopyProperties, true) | kStaticProperties);
 }
 
+
 template<class A, class U>
 ConstFstImpl<A, U> *ConstFstImpl<A, U>::Read(istream &strm,
                                              const FstReadOptions &opts) {
   ConstFstImpl<A, U> *impl = new ConstFstImpl<A, U>;
   FstHeader hdr;
-  if (!impl->ReadHeader(strm, opts, kMinFileVersion, &hdr))
+  if (!impl->ReadHeader(strm, opts, kMinFileVersion, &hdr)) {
+    delete impl;
     return 0;
+  }
   impl->start_ = hdr.Start();
   impl->nstates_ = hdr.NumStates();
   impl->narcs_ = hdr.NumArcs();
   impl->states_ = new State[impl->nstates_];
   impl->arcs_ = new A[impl->narcs_];
 
-  char c;
-  for (int i = 0; i < kFileAlign && strm.tellg() % kFileAlign; ++i)
-    strm.read(&c, 1);
-  // TODO: memory map this
+  // Ensures compatibility
+  if (hdr.Version() == kAlignedFileVersion)
+    hdr.SetFlags(hdr.GetFlags() | FstHeader::IS_ALIGNED);
+
+  if ((hdr.GetFlags() & FstHeader::IS_ALIGNED) &&
+      !AlignInput(strm, kFileAlign)) {
+    LOG(ERROR) << "ConstFst::Read: Alignment failed: " << opts.source;
+    delete impl;
+    return 0;
+  }
   size_t b = impl->nstates_ * sizeof(typename ConstFstImpl<A, U>::State);
   strm.read(reinterpret_cast<char *>(impl->states_), b);
   if (!strm) {
     LOG(ERROR) << "ConstFst::Read: Read failed: " << opts.source;
+    delete impl;
+    return 0;
+  }
+  if ((hdr.GetFlags() & FstHeader::IS_ALIGNED) &&
+      !AlignInput(strm, kFileAlign)) {
+    LOG(ERROR) << "ConstFst::Read: Alignment failed: " << opts.source;
+    delete impl;
     return 0;
   }
-  // TODO: memory map this
-  for (int i = 0; i < kFileAlign && strm.tellg() % kFileAlign; ++i)
-    strm.read(&c, 1);
   b = impl->narcs_ * sizeof(A);
   strm.read(reinterpret_cast<char *>(impl->arcs_), b);
   if (!strm) {
     LOG(ERROR) << "ConstFst::Read: Read failed: " << opts.source;
+    delete impl;
     return 0;
   }
   return impl;
 }
 
-template<class A, class U>
-bool ConstFstImpl<A, U>::Write(ostream &strm,
-                               const FstWriteOptions &opts) const {
-  FstHeader hdr;
-  hdr.SetStart(start_);
-  hdr.SetNumStates(nstates_);
-  hdr.SetNumArcs(narcs_);
-  WriteHeader(strm, opts, kFileVersion, &hdr);
-
-  for (int i = 0; i < kFileAlign && strm.tellp() % kFileAlign; ++i)
-    strm.write("", 1);
-  strm.write(reinterpret_cast<char *>(states_), nstates_ * sizeof(State));
-  for (int i = 0; i < kFileAlign && strm.tellp() % kFileAlign; ++i)
-    strm.write("", 1);
-  strm.write(reinterpret_cast<char *>(arcs_), narcs_ * sizeof(A));
-
-  strm.flush();
-  if (!strm) {
-    LOG(ERROR) << "ConstFst::Write: Write failed: " << opts.source;
-    return false;
-  }
-  return true;
-}
-
-
 // Simple concrete immutable FST.  This class attaches interface to
 // implementation and handles reference counting, delegating most
 // methods to ImplToExpandedFst. The unsigned type U is used to
@@ -294,6 +290,18 @@ class ConstFst : public ImplToExpandedFst< ConstFstImpl<A, U> > {
     return impl ? new ConstFst<A, U>(impl) : 0;
   }
 
+  virtual bool Write(ostream &strm, const FstWriteOptions &opts) const {
+    return WriteFst(*this, strm, opts);
+  }
+
+  virtual bool Write(const string &filename) const {
+    return Fst<A>::WriteFile(filename);
+  }
+
+  template <class F>
+  static bool WriteFst(const F &fst, ostream &strm,
+                       const FstWriteOptions &opts);
+
   virtual void InitStateIterator(StateIteratorData<Arc> *data) const {
     GetImpl()->InitStateIterator(data);
   }
@@ -315,6 +323,93 @@ class ConstFst : public ImplToExpandedFst< ConstFstImpl<A, U> > {
   void operator=(const ConstFst<A, U> &fst);  // disallow
 };
 
+// Writes Fst in Const format, potentially with a pass over the machine
+// before writing to compute number of states and arcs.
+//
+template <class A, class U>
+template <class F>
+bool ConstFst<A, U>::WriteFst(const F &fst, ostream &strm,
+                              const FstWriteOptions &opts) {
+  static const int kFileVersion = 2;
+  static const int kAlignedFileVersion = 1;
+  static const int kFileAlign = 16;
+  int file_version = opts.align ? kAlignedFileVersion : kFileVersion;
+  size_t num_arcs = -1, num_states = -1;
+  size_t start_offset = 0;
+  bool update_header = true;
+  if (fst.Type() == ConstFst<A, U>().Type()) {
+    const ConstFst<A, U> *const_fst =
+        reinterpret_cast<const ConstFst<A, U> *>(&fst);
+    num_arcs = const_fst->GetImpl()->narcs_;
+    num_states = const_fst->GetImpl()->nstates_;
+    update_header = false;
+  } else if ((start_offset = strm.tellp()) == -1) {
+    // precompute values needed for header when we cannot seek to rewrite it.
+    for (StateIterator<F> siter(fst); !siter.Done(); siter.Next()) {
+      num_arcs += fst.NumArcs(siter.Value());
+      num_states++;
+    }
+    update_header = false;
+  }
+  FstHeader hdr;
+  hdr.SetStart(fst.Start());
+  hdr.SetNumStates(num_states);
+  hdr.SetNumArcs(num_arcs);
+  string type = "const";
+  if (sizeof(U) != sizeof(uint32)) {
+    string size;
+    Int64ToStr(8 * sizeof(U), &size);
+    type += size;
+  }
+  FstImpl<A>::WriteFstHeader(fst, strm, opts, file_version, type, &hdr);
+  if (opts.align && !AlignOutput(strm, kFileAlign)) {
+    LOG(ERROR) << "Could not align file during write after header";
+    return false;
+  }
+  size_t pos = 0, states = 0;
+  typename ConstFstImpl<A, U>::State state;
+  for (StateIterator<F> siter(fst); !siter.Done(); siter.Next()) {
+    state.final = fst.Final(siter.Value());
+    state.pos = pos;
+    state.narcs = fst.NumArcs(siter.Value());
+    state.niepsilons = fst.NumInputEpsilons(siter.Value());
+    state.noepsilons = fst.NumOutputEpsilons(siter.Value());
+    strm.write(reinterpret_cast<const char *>(&state), sizeof(state));
+    pos += state.narcs;
+    states++;
+  }
+  hdr.SetNumStates(states);
+  hdr.SetNumArcs(pos);
+  if (opts.align && !AlignOutput(strm, kFileAlign)) {
+    LOG(ERROR) << "Could not align file during write after writing states";
+  }
+  for (StateIterator<F> siter(fst); !siter.Done(); siter.Next()) {
+    StateId s = siter.Value();
+    for (ArcIterator<F> aiter(fst, s); !aiter.Done(); aiter.Next()) {
+      const A &arc = aiter.Value();
+      strm.write(reinterpret_cast<const char *>(&arc), sizeof(arc));
+    }
+  }
+  strm.flush();
+  if (!strm) {
+    LOG(ERROR) << "WriteAsVectorFst write failed: " << opts.source;
+    return false;
+  }
+  if (update_header) {
+    return FstImpl<A>::UpdateFstHeader(fst, strm, opts, file_version, type,
+                                       &hdr, start_offset);
+  } else {
+    if (hdr.NumStates() != num_states) {
+      LOG(ERROR) << "Inconsistent number of states observed during write";
+      return false;
+    }
+    if (hdr.NumArcs() != num_arcs) {
+      LOG(ERROR) << "Inconsistent number of arcs observed during write";
+      return false;
+    }
+  }
+  return true;
+}
 
 // Specialization for ConstFst; see generic version in fst.h
 // for sample usage (but use the ConstFst type!). This version
@@ -384,6 +479,6 @@ class ArcIterator< ConstFst<A, U> > {
 // A useful alias when using StdArc.
 typedef ConstFst<StdArc> StdConstFst;
 
-}  // namespace fst;
+}  // namespace fst
 
 #endif  // FST_LIB_CONST_FST_H__
