@@ -26,29 +26,34 @@
 #include <climits>
 #include <sstream>
 #include <string>
+
 #include <fst/util.h>
 #include <fst/weight.h>
 
+
 namespace fst {
 
 // numeric limits class
 template <class T>
 class FloatLimits {
  public:
-  static const T kPosInfinity;
-  static const T kNegInfinity;
-  static const T kNumberBad;
+  static const T PosInfinity() {
+    static const T pos_infinity = numeric_limits<T>::infinity();
+    return pos_infinity;
+  }
+
+  static const T NegInfinity() {
+    static const T neg_infinity = -PosInfinity();
+    return neg_infinity;
+  }
+
+  static const T NumberBad() {
+    static const T number_bad = numeric_limits<T>::quiet_NaN();
+    return number_bad;
+  }
+
 };
 
-template <class T>
-const T FloatLimits<T>::kPosInfinity = numeric_limits<T>::infinity();
-
-template <class T>
-const T FloatLimits<T>::kNegInfinity = -FloatLimits<T>::kPosInfinity;
-
-template <class T>
-const T FloatLimits<T>::kNumberBad = numeric_limits<T>::quiet_NaN();
-
 // weight class to be templated on floating-points types
 template <class T = float>
 class FloatWeightTpl {
@@ -85,6 +90,8 @@ class FloatWeightTpl {
   const T &Value() const { return value_; }
 
  protected:
+  void SetValue(const T &f) { value_ = f; }
+
   inline static string GetPrecisionString() {
     int64 size = sizeof(T);
     if (size == sizeof(float)) return "";
@@ -95,10 +102,11 @@ class FloatWeightTpl {
     return result;
   }
 
+ private:
   T value_;
 };
 
-// Single precision FloatWeight
+// Single-precision float weight
 typedef FloatWeightTpl<float> FloatWeight;
 
 template <class T>
@@ -146,9 +154,9 @@ inline bool ApproxEqual(const FloatWeightTpl<T> &w1,
 
 template <class T>
 inline ostream &operator<<(ostream &strm, const FloatWeightTpl<T> &w) {
-  if (w.Value() == FloatLimits<T>::kPosInfinity)
+  if (w.Value() == FloatLimits<T>::PosInfinity())
     return strm << "Infinity";
-  else if (w.Value() == FloatLimits<T>::kNegInfinity)
+  else if (w.Value() == FloatLimits<T>::NegInfinity())
     return strm << "-Infinity";
   else if (w.Value() != w.Value())   // Fails for NaN
     return strm << "BadNumber";
@@ -161,9 +169,9 @@ inline istream &operator>>(istream &strm, FloatWeightTpl<T> &w) {
   string s;
   strm >> s;
   if (s == "Infinity") {
-    w = FloatWeightTpl<T>(FloatLimits<T>::kPosInfinity);
+    w = FloatWeightTpl<T>(FloatLimits<T>::PosInfinity());
   } else if (s == "-Infinity") {
-    w = FloatWeightTpl<T>(FloatLimits<T>::kNegInfinity);
+    w = FloatWeightTpl<T>(FloatLimits<T>::NegInfinity());
   } else {
     char *p;
     T f = strtod(s.c_str(), &p);
@@ -191,11 +199,14 @@ class TropicalWeightTpl : public FloatWeightTpl<T> {
   TropicalWeightTpl(const TropicalWeightTpl<T> &w) : FloatWeightTpl<T>(w) {}
 
   static const TropicalWeightTpl<T> Zero() {
-    return TropicalWeightTpl<T>(FloatLimits<T>::kPosInfinity); }
+    return TropicalWeightTpl<T>(FloatLimits<T>::PosInfinity()); }
 
   static const TropicalWeightTpl<T> One() {
     return TropicalWeightTpl<T>(0.0F); }
 
+  static const TropicalWeightTpl<T> NoWeight() {
+    return TropicalWeightTpl<T>(FloatLimits<T>::NumberBad()); }
+
   static const string &Type() {
     static const string type = "tropical" +
         FloatWeightTpl<T>::GetPrecisionString();
@@ -204,12 +215,12 @@ class TropicalWeightTpl : public FloatWeightTpl<T> {
 
   bool Member() const {
     // First part fails for IEEE NaN
-    return Value() == Value() && Value() != FloatLimits<T>::kNegInfinity;
+    return Value() == Value() && Value() != FloatLimits<T>::NegInfinity();
   }
 
   TropicalWeightTpl<T> Quantize(float delta = kDelta) const {
-    if (Value() == FloatLimits<T>::kNegInfinity ||
-        Value() == FloatLimits<T>::kPosInfinity ||
+    if (Value() == FloatLimits<T>::NegInfinity() ||
+        Value() == FloatLimits<T>::PosInfinity() ||
         Value() != Value())
       return *this;
     else
@@ -224,12 +235,14 @@ class TropicalWeightTpl : public FloatWeightTpl<T> {
   }
 };
 
-// Single precision TropicalWeight
+// Single precision tropical weight
 typedef TropicalWeightTpl<float> TropicalWeight;
 
 template <class T>
 inline TropicalWeightTpl<T> Plus(const TropicalWeightTpl<T> &w1,
                                  const TropicalWeightTpl<T> &w2) {
+  if (!w1.Member() || !w2.Member())
+    return TropicalWeightTpl<T>::NoWeight();
   return w1.Value() < w2.Value() ? w1 : w2;
 }
 
@@ -246,10 +259,12 @@ inline TropicalWeightTpl<double> Plus(const TropicalWeightTpl<double> &w1,
 template <class T>
 inline TropicalWeightTpl<T> Times(const TropicalWeightTpl<T> &w1,
                                   const TropicalWeightTpl<T> &w2) {
+  if (!w1.Member() || !w2.Member())
+    return TropicalWeightTpl<T>::NoWeight();
   T f1 = w1.Value(), f2 = w2.Value();
-  if (f1 == FloatLimits<T>::kPosInfinity)
+  if (f1 == FloatLimits<T>::PosInfinity())
     return w1;
-  else if (f2 == FloatLimits<T>::kPosInfinity)
+  else if (f2 == FloatLimits<T>::PosInfinity())
     return w2;
   else
     return TropicalWeightTpl<T>(f1 + f2);
@@ -269,11 +284,13 @@ template <class T>
 inline TropicalWeightTpl<T> Divide(const TropicalWeightTpl<T> &w1,
                                    const TropicalWeightTpl<T> &w2,
                                    DivideType typ = DIVIDE_ANY) {
+  if (!w1.Member() || !w2.Member())
+    return TropicalWeightTpl<T>::NoWeight();
   T f1 = w1.Value(), f2 = w2.Value();
-  if (f2 == FloatLimits<T>::kPosInfinity)
-    return FloatLimits<T>::kNumberBad;
-  else if (f1 == FloatLimits<T>::kPosInfinity)
-    return FloatLimits<T>::kPosInfinity;
+  if (f2 == FloatLimits<T>::PosInfinity())
+    return FloatLimits<T>::NumberBad();
+  else if (f1 == FloatLimits<T>::PosInfinity())
+    return FloatLimits<T>::PosInfinity();
   else
     return TropicalWeightTpl<T>(f1 - f2);
 }
@@ -306,13 +323,16 @@ class LogWeightTpl : public FloatWeightTpl<T> {
   LogWeightTpl(const LogWeightTpl<T> &w) : FloatWeightTpl<T>(w) {}
 
   static const LogWeightTpl<T> Zero() {
-    return LogWeightTpl<T>(FloatLimits<T>::kPosInfinity);
+    return LogWeightTpl<T>(FloatLimits<T>::PosInfinity());
   }
 
   static const LogWeightTpl<T> One() {
     return LogWeightTpl<T>(0.0F);
   }
 
+  static const LogWeightTpl<T> NoWeight() {
+    return LogWeightTpl<T>(FloatLimits<T>::NumberBad()); }
+
   static const string &Type() {
     static const string type = "log" + FloatWeightTpl<T>::GetPrecisionString();
     return type;
@@ -320,12 +340,12 @@ class LogWeightTpl : public FloatWeightTpl<T> {
 
   bool Member() const {
     // First part fails for IEEE NaN
-    return Value() == Value() && Value() != FloatLimits<T>::kNegInfinity;
+    return Value() == Value() && Value() != FloatLimits<T>::NegInfinity();
   }
 
   LogWeightTpl<T> Quantize(float delta = kDelta) const {
-    if (Value() == FloatLimits<T>::kNegInfinity ||
-        Value() == FloatLimits<T>::kPosInfinity ||
+    if (Value() == FloatLimits<T>::NegInfinity() ||
+        Value() == FloatLimits<T>::PosInfinity() ||
         Value() != Value())
       return *this;
     else
@@ -339,8 +359,10 @@ class LogWeightTpl : public FloatWeightTpl<T> {
   }
 };
 
-// Single precision LogWeight
+// Single-precision log weight
 typedef LogWeightTpl<float> LogWeight;
+// Double-precision log weight
+typedef LogWeightTpl<double> Log64Weight;
 
 template <class T>
 inline T LogExp(T x) { return log(1.0F + exp(-x)); }
@@ -349,9 +371,9 @@ template <class T>
 inline LogWeightTpl<T> Plus(const LogWeightTpl<T> &w1,
                             const LogWeightTpl<T> &w2) {
   T f1 = w1.Value(), f2 = w2.Value();
-  if (f1 == FloatLimits<T>::kPosInfinity)
+  if (f1 == FloatLimits<T>::PosInfinity())
     return w2;
-  else if (f2 == FloatLimits<T>::kPosInfinity)
+  else if (f2 == FloatLimits<T>::PosInfinity())
     return w1;
   else if (f1 > f2)
     return LogWeightTpl<T>(f2 - LogExp(f1 - f2));
@@ -372,10 +394,12 @@ inline LogWeightTpl<double> Plus(const LogWeightTpl<double> &w1,
 template <class T>
 inline LogWeightTpl<T> Times(const LogWeightTpl<T> &w1,
                              const LogWeightTpl<T> &w2) {
+  if (!w1.Member() || !w2.Member())
+    return LogWeightTpl<T>::NoWeight();
   T f1 = w1.Value(), f2 = w2.Value();
-  if (f1 == FloatLimits<T>::kPosInfinity)
+  if (f1 == FloatLimits<T>::PosInfinity())
     return w1;
-  else if (f2 == FloatLimits<T>::kPosInfinity)
+  else if (f2 == FloatLimits<T>::PosInfinity())
     return w2;
   else
     return LogWeightTpl<T>(f1 + f2);
@@ -395,11 +419,13 @@ template <class T>
 inline LogWeightTpl<T> Divide(const LogWeightTpl<T> &w1,
                               const LogWeightTpl<T> &w2,
                               DivideType typ = DIVIDE_ANY) {
+  if (!w1.Member() || !w2.Member())
+    return LogWeightTpl<T>::NoWeight();
   T f1 = w1.Value(), f2 = w2.Value();
-  if (f2 == FloatLimits<T>::kPosInfinity)
-    return FloatLimits<T>::kNumberBad;
-  else if (f1 == FloatLimits<T>::kPosInfinity)
-    return FloatLimits<T>::kPosInfinity;
+  if (f2 == FloatLimits<T>::PosInfinity())
+    return FloatLimits<T>::NumberBad();
+  else if (f1 == FloatLimits<T>::PosInfinity())
+    return FloatLimits<T>::PosInfinity();
   else
     return LogWeightTpl<T>(f1 - f2);
 }
@@ -431,13 +457,16 @@ class MinMaxWeightTpl : public FloatWeightTpl<T> {
   MinMaxWeightTpl(const MinMaxWeightTpl<T> &w) : FloatWeightTpl<T>(w) {}
 
   static const MinMaxWeightTpl<T> Zero() {
-    return MinMaxWeightTpl<T>(FloatLimits<T>::kPosInfinity);
+    return MinMaxWeightTpl<T>(FloatLimits<T>::PosInfinity());
   }
 
   static const MinMaxWeightTpl<T> One() {
-    return MinMaxWeightTpl<T>(FloatLimits<T>::kNegInfinity);
+    return MinMaxWeightTpl<T>(FloatLimits<T>::NegInfinity());
   }
 
+  static const MinMaxWeightTpl<T> NoWeight() {
+    return MinMaxWeightTpl<T>(FloatLimits<T>::NumberBad()); }
+
   static const string &Type() {
     static const string type = "minmax" +
         FloatWeightTpl<T>::GetPrecisionString();
@@ -451,8 +480,8 @@ class MinMaxWeightTpl : public FloatWeightTpl<T> {
 
   MinMaxWeightTpl<T> Quantize(float delta = kDelta) const {
     // If one of infinities, or a NaN
-    if (Value() == FloatLimits<T>::kNegInfinity ||
-        Value() == FloatLimits<T>::kPosInfinity ||
+    if (Value() == FloatLimits<T>::NegInfinity() ||
+        Value() == FloatLimits<T>::PosInfinity() ||
         Value() != Value())
       return *this;
     else
@@ -466,13 +495,15 @@ class MinMaxWeightTpl : public FloatWeightTpl<T> {
   }
 };
 
-// Single precision MinMaxWeight
+// Single-precision min-max weight
 typedef MinMaxWeightTpl<float> MinMaxWeight;
 
 // Min
 template <class T>
 inline MinMaxWeightTpl<T> Plus(
     const MinMaxWeightTpl<T> &w1, const MinMaxWeightTpl<T> &w2) {
+  if (!w1.Member() || !w2.Member())
+    return MinMaxWeightTpl<T>::NoWeight();
   return w1.Value() < w2.Value() ? w1 : w2;
 }
 
@@ -490,6 +521,8 @@ inline MinMaxWeightTpl<double> Plus(
 template <class T>
 inline MinMaxWeightTpl<T> Times(
     const MinMaxWeightTpl<T> &w1, const MinMaxWeightTpl<T> &w2) {
+  if (!w1.Member() || !w2.Member())
+    return MinMaxWeightTpl<T>::NoWeight();
   return w1.Value() >= w2.Value() ? w1 : w2;
 }
 
@@ -508,8 +541,10 @@ template <class T>
 inline MinMaxWeightTpl<T> Divide(const MinMaxWeightTpl<T> &w1,
                                  const MinMaxWeightTpl<T> &w2,
                                  DivideType typ = DIVIDE_ANY) {
+  if (!w1.Member() || !w2.Member())
+    return MinMaxWeightTpl<T>::NoWeight();
   // min(w1, x) = w2, w1 >= w2 => min(w1, x) = w2, x = w2
-  return w1.Value() >= w2.Value() ? w1 : FloatLimits<T>::kNumberBad;
+  return w1.Value() >= w2.Value() ? w1 : FloatLimits<T>::NumberBad();
 }
 
 inline MinMaxWeightTpl<float> Divide(const MinMaxWeightTpl<float> &w1,
@@ -524,7 +559,43 @@ inline MinMaxWeightTpl<double> Divide(const MinMaxWeightTpl<double> &w1,
   return Divide<double>(w1, w2, typ);
 }
 
+//
+// WEIGHT CONVERTER SPECIALIZATIONS.
+//
 
-}  // namespace fst;
+// Convert to tropical
+template <>
+struct WeightConvert<LogWeight, TropicalWeight> {
+  TropicalWeight operator()(LogWeight w) const { return w.Value(); }
+};
+
+template <>
+struct WeightConvert<Log64Weight, TropicalWeight> {
+  TropicalWeight operator()(Log64Weight w) const { return w.Value(); }
+};
+
+// Convert to log
+template <>
+struct WeightConvert<TropicalWeight, LogWeight> {
+  LogWeight operator()(TropicalWeight w) const { return w.Value(); }
+};
+
+template <>
+struct WeightConvert<Log64Weight, LogWeight> {
+  LogWeight operator()(Log64Weight w) const { return w.Value(); }
+};
+
+// Convert to log64
+template <>
+struct WeightConvert<TropicalWeight, Log64Weight> {
+  Log64Weight operator()(TropicalWeight w) const { return w.Value(); }
+};
+
+template <>
+struct WeightConvert<LogWeight, Log64Weight> {
+  Log64Weight operator()(LogWeight w) const { return w.Value(); }
+};
+
+}  // namespace fst
 
 #endif  // FST_LIB_FLOAT_WEIGHT_H__
