@@ -28,6 +28,7 @@
 #include <fst/label-reachable.h>
 #include <fst/matcher.h>
 
+
 DECLARE_string(save_relabel_ipairs);
 DECLARE_string(save_relabel_opairs);
 
@@ -306,81 +307,7 @@ class ArcLookAheadMatcher
 
   // Checks if there is a matching (possibly super-final) transition
   // at (s_, s).
-  bool LookAheadFst(const Fst<Arc> &fst, StateId s) {
-    if (&fst != lfst_)
-      InitLookAheadFst(fst);
-
-    bool ret = false;
-    ssize_t nprefix = 0;
-    if (F & kLookAheadWeight)
-      SetLookAheadWeight(Weight::Zero());
-    if (F & kLookAheadPrefix)
-      ClearLookAheadPrefix();
-    if (fst_.Final(s_) != Weight::Zero() &&
-        lfst_->Final(s) != Weight::Zero()) {
-      if (!(F & (kLookAheadWeight | kLookAheadPrefix)))
-        return true;
-      ++nprefix;
-      if (F & kLookAheadWeight)
-        SetLookAheadWeight(Plus(LookAheadWeight(),
-                                Times(fst_.Final(s_), lfst_->Final(s))));
-      ret = true;
-    }
-    if (matcher_.Find(kNoLabel)) {
-      if (!(F & (kLookAheadWeight | kLookAheadPrefix)))
-        return true;
-      ++nprefix;
-      if (F & kLookAheadWeight)
-        for (; !matcher_.Done(); matcher_.Next())
-          SetLookAheadWeight(Plus(LookAheadWeight(), matcher_.Value().weight));
-      ret = true;
-    }
-    for (ArcIterator< Fst<Arc> > aiter(*lfst_, s);
-         !aiter.Done();
-         aiter.Next()) {
-      const Arc &arc = aiter.Value();
-      Label label = kNoLabel;
-      switch (matcher_.Type(false)) {
-        case MATCH_INPUT:
-          label = arc.olabel;
-          break;
-        case MATCH_OUTPUT:
-          label = arc.ilabel;
-          break;
-        default:
-          LOG(FATAL) << "ArcLookAheadMatcher::LookAheadFst: bad match type";
-      }
-      if (label == 0) {
-        if (!(F & (kLookAheadWeight | kLookAheadPrefix)))
-          return true;
-        if (!(F & kLookAheadNonEpsilonPrefix))
-          ++nprefix;
-        if (F & kLookAheadWeight)
-          SetLookAheadWeight(Plus(LookAheadWeight(), arc.weight));
-        ret = true;
-      } else if (matcher_.Find(label)) {
-        if (!(F & (kLookAheadWeight | kLookAheadPrefix)))
-          return true;
-        for (; !matcher_.Done(); matcher_.Next()) {
-          ++nprefix;
-          if (F & kLookAheadWeight)
-            SetLookAheadWeight(Plus(LookAheadWeight(),
-                                    Times(arc.weight,
-                                          matcher_.Value().weight)));
-          if ((F & kLookAheadPrefix) && nprefix == 1)
-            SetLookAheadPrefix(arc);
-        }
-        ret = true;
-      }
-    }
-    if (F & kLookAheadPrefix) {
-      if (nprefix == 1)
-        SetLookAheadWeight(Weight::One());  // Avoids double counting.
-      else
-        ClearLookAheadPrefix();
-    }
-    return ret;
-  }
+  bool LookAheadFst(const Fst<Arc> &fst, StateId s);
 
   void InitLookAheadFst(const Fst<Arc>& fst, bool copy = false) {
     lfst_ = &fst;
@@ -407,6 +334,85 @@ class ArcLookAheadMatcher
   StateId s_;              // Matcher state
 };
 
+template <class M, uint32 F>
+bool ArcLookAheadMatcher<M, F>::LookAheadFst(const Fst<Arc> &fst, StateId s) {
+  if (&fst != lfst_)
+    InitLookAheadFst(fst);
+
+  bool ret = false;
+  ssize_t nprefix = 0;
+  if (F & kLookAheadWeight)
+    SetLookAheadWeight(Weight::Zero());
+  if (F & kLookAheadPrefix)
+    ClearLookAheadPrefix();
+  if (fst_.Final(s_) != Weight::Zero() &&
+      lfst_->Final(s) != Weight::Zero()) {
+    if (!(F & (kLookAheadWeight | kLookAheadPrefix)))
+      return true;
+    ++nprefix;
+    if (F & kLookAheadWeight)
+      SetLookAheadWeight(Plus(LookAheadWeight(),
+                              Times(fst_.Final(s_), lfst_->Final(s))));
+    ret = true;
+  }
+  if (matcher_.Find(kNoLabel)) {
+    if (!(F & (kLookAheadWeight | kLookAheadPrefix)))
+      return true;
+    ++nprefix;
+    if (F & kLookAheadWeight)
+      for (; !matcher_.Done(); matcher_.Next())
+        SetLookAheadWeight(Plus(LookAheadWeight(), matcher_.Value().weight));
+    ret = true;
+  }
+  for (ArcIterator< Fst<Arc> > aiter(*lfst_, s);
+       !aiter.Done();
+       aiter.Next()) {
+    const Arc &arc = aiter.Value();
+    Label label = kNoLabel;
+    switch (matcher_.Type(false)) {
+      case MATCH_INPUT:
+        label = arc.olabel;
+        break;
+      case MATCH_OUTPUT:
+        label = arc.ilabel;
+        break;
+      default:
+        FSTERROR() << "ArcLookAheadMatcher::LookAheadFst: bad match type";
+        return true;
+    }
+    if (label == 0) {
+      if (!(F & (kLookAheadWeight | kLookAheadPrefix)))
+        return true;
+      if (!(F & kLookAheadNonEpsilonPrefix))
+        ++nprefix;
+      if (F & kLookAheadWeight)
+        SetLookAheadWeight(Plus(LookAheadWeight(), arc.weight));
+      ret = true;
+    } else if (matcher_.Find(label)) {
+      if (!(F & (kLookAheadWeight | kLookAheadPrefix)))
+        return true;
+      for (; !matcher_.Done(); matcher_.Next()) {
+        ++nprefix;
+        if (F & kLookAheadWeight)
+          SetLookAheadWeight(Plus(LookAheadWeight(),
+                                  Times(arc.weight,
+                                        matcher_.Value().weight)));
+        if ((F & kLookAheadPrefix) && nprefix == 1)
+          SetLookAheadPrefix(arc);
+      }
+      ret = true;
+    }
+  }
+  if (F & kLookAheadPrefix) {
+    if (nprefix == 1)
+      SetLookAheadWeight(Weight::One());  // Avoids double counting.
+    else
+      ClearLookAheadPrefix();
+  }
+  return ret;
+}
+
+
 // Template argument F accepts flags to control behavior.
 // It must include precisely one of KInputLookAheadMatcher or
 // KOutputLookAheadMatcher.
@@ -434,16 +440,20 @@ class LabelLookAheadMatcher
       : matcher_(fst, match_type),
         lfst_(0),
         label_reachable_(0),
-        s_(kNoStateId) {
-    CHECK(F & (kInputLookAheadMatcher | kOutputLookAheadMatcher));
+        s_(kNoStateId),
+        error_(false) {
+    if (!(F & (kInputLookAheadMatcher | kOutputLookAheadMatcher))) {
+      FSTERROR() << "LabelLookaheadMatcher: bad matcher flags: " << F;
+      error_ = true;
+    }
     bool reach_input = match_type == MATCH_INPUT;
     if (data) {
       if (reach_input == data->ReachInput())
         label_reachable_ = new LabelReachable<Arc, S>(data, s);
     } else if ((reach_input && (F & kInputLookAheadMatcher)) ||
                (!reach_input && (F & kOutputLookAheadMatcher))) {
-      label_reachable_ = new LabelReachable<Arc, S>(fst, reach_input, s,
-                                                    F & kLookAheadKeepRelabelData);
+      label_reachable_ = new LabelReachable<Arc, S>(
+          fst, reach_input, s, F & kLookAheadKeepRelabelData);
     }
   }
 
@@ -454,9 +464,8 @@ class LabelLookAheadMatcher
         label_reachable_(
             lmatcher.label_reachable_ ?
             new LabelReachable<Arc, S>(*lmatcher.label_reachable_) : 0),
-        s_(kNoStateId) {
-    CHECK(F & (kInputLookAheadMatcher | kOutputLookAheadMatcher));
-  }
+        s_(kNoStateId),
+        error_(lmatcher.error_) {}
 
   ~LabelLookAheadMatcher() {
     delete label_reachable_;
@@ -489,7 +498,13 @@ class LabelLookAheadMatcher
   const Arc& Value() const { return matcher_.Value(); }
   void Next() { matcher_.Next(); }
   const FST &GetFst() const { return matcher_.GetFst(); }
-  uint64 Properties(uint64 props) const { return matcher_.Properties(props); }
+
+  uint64 Properties(uint64 inprops) const {
+    uint64 outprops = matcher_.Properties(inprops);
+    if (error_ || (label_reachable_ && label_reachable_->Error()))
+      outprops |= kError;
+    return outprops;
+  }
 
   uint32 Flags() const {
     if (label_reachable_ && label_reachable_->GetData()->ReachInput())
@@ -507,6 +522,9 @@ class LabelLookAheadMatcher
 
   // Look-ahead methods.
   bool LookAheadLabel(Label label) const {
+    if (label == 0)
+      return true;
+
     if (label_reachable_) {
       if (!reach_set_state_) {
         label_reachable_->SetState(s_);
@@ -521,46 +539,7 @@ class LabelLookAheadMatcher
   // Checks if there is a matching (possibly super-final) transition
   // at (s_, s).
   template <class L>
-  bool LookAheadFst(const L &fst, StateId s) {
-    if (static_cast<const Fst<Arc> *>(&fst) != lfst_)
-      InitLookAheadFst(fst);
-
-    SetLookAheadWeight(Weight::One());
-    ClearLookAheadPrefix();
-
-    if (!label_reachable_)
-      return true;
-
-    label_reachable_->SetState(s_, s);
-    reach_set_state_ = true;
-
-    bool compute_weight = F & kLookAheadWeight;
-    bool compute_prefix = F & kLookAheadPrefix;
-
-    bool reach_input = Type(false) == MATCH_OUTPUT;
-    ArcIterator<L> aiter(fst, s);
-    bool reach_arc = label_reachable_->Reach(&aiter, 0, NumArcs(*lfst_, s),
-                                             reach_input, compute_weight);
-    if (reach_arc) {
-      ssize_t begin = label_reachable_->ReachBegin();
-      ssize_t end = label_reachable_->ReachEnd();
-      if (compute_prefix && end - begin == 1) {
-        aiter.Seek(begin);
-        SetLookAheadPrefix(aiter.Value());
-        compute_weight = false;
-      } else if (compute_weight) {
-        SetLookAheadWeight(label_reachable_->ReachWeight());
-      }
-    }
-    Weight lfinal = Final(*lfst_, s);
-    bool reach_final = lfinal != Weight::Zero() &&
-        label_reachable_->ReachFinal();
-    if (reach_final && compute_weight)
-           SetLookAheadWeight(reach_arc ?
-                              Plus(LookAheadWeight(), lfinal) : lfinal);
-
-    return reach_arc || reach_final;
-  }
+  bool LookAheadFst(const L &fst, StateId s);
 
   void InitLookAheadFst(const Fst<Arc>& fst, bool copy = false) {
     lfst_ = &fst;
@@ -596,8 +575,52 @@ class LabelLookAheadMatcher
   StateId s_;                                // Matcher state
   bool match_set_state_;                     // matcher_.SetState called?
   mutable bool reach_set_state_;             // reachable_.SetState called?
+  bool error_;
 };
 
+template <class M, uint32 F, class S>
+template <class L> inline
+bool LabelLookAheadMatcher<M, F, S>::LookAheadFst(const L &fst, StateId s) {
+  if (static_cast<const Fst<Arc> *>(&fst) != lfst_)
+    InitLookAheadFst(fst);
+
+  SetLookAheadWeight(Weight::One());
+  ClearLookAheadPrefix();
+
+  if (!label_reachable_)
+    return true;
+
+  label_reachable_->SetState(s_, s);
+  reach_set_state_ = true;
+
+  bool compute_weight = F & kLookAheadWeight;
+  bool compute_prefix = F & kLookAheadPrefix;
+
+  bool reach_input = Type(false) == MATCH_OUTPUT;
+  ArcIterator<L> aiter(fst, s);
+  bool reach_arc = label_reachable_->Reach(&aiter, 0,
+                                           internal::NumArcs(*lfst_, s),
+                                           reach_input, compute_weight);
+  Weight lfinal = internal::Final(*lfst_, s);
+  bool reach_final = lfinal != Weight::Zero() && label_reachable_->ReachFinal();
+  if (reach_arc) {
+    ssize_t begin = label_reachable_->ReachBegin();
+    ssize_t end = label_reachable_->ReachEnd();
+    if (compute_prefix && end - begin == 1 && !reach_final) {
+      aiter.Seek(begin);
+      SetLookAheadPrefix(aiter.Value());
+      compute_weight = false;
+    } else if (compute_weight) {
+      SetLookAheadWeight(label_reachable_->ReachWeight());
+    }
+  }
+  if (reach_final && compute_weight)
+    SetLookAheadWeight(reach_arc ?
+                       Plus(LookAheadWeight(), lfinal) : lfinal);
+
+  return reach_arc || reach_final;
+}
+
 
 // Label-lookahead relabeling class.
 template <class A>
@@ -609,43 +632,7 @@ class LabelLookAheadRelabeler {
 
   // Relabels matcher Fst - initialization function object.
   template <typename I>
-  LabelLookAheadRelabeler(I **impl) {
-    Fst<A> &fst = (*impl)->GetFst();
-    D *data = (*impl)->GetAddOn();
-    const string name = (*impl)->Type();
-    bool is_mutable = fst.Properties(kMutable, false);
-    MutableFst<A> *mfst = 0;
-    if (is_mutable) {
-      mfst = static_cast<MutableFst<A> *>(&fst);
-    } else {
-      mfst = new VectorFst<A>(fst);
-      data->IncrRefCount();
-      delete *impl;
-    }
-    if (data->First()) {  // reach_input
-      LabelReachable<A> reachable(data->First());
-      reachable.Relabel(mfst, true);
-      if (!FLAGS_save_relabel_ipairs.empty()) {
-        vector<pair<Label, Label> > pairs;
-        reachable.RelabelPairs(&pairs);
-        WriteLabelPairs(FLAGS_save_relabel_ipairs, pairs);
-      }
-    } else {
-      LabelReachable<A> reachable(data->Second());
-      reachable.Relabel(mfst, false);
-      if (!FLAGS_save_relabel_opairs.empty()) {
-        vector<pair<Label, Label> > pairs;
-        reachable.RelabelPairs(&pairs);
-        WriteLabelPairs(FLAGS_save_relabel_opairs, pairs);
-      }
-    }
-    if (!is_mutable) {
-      *impl = new I(*mfst, name);
-      (*impl)->SetAddOn(data);
-      delete mfst;
-      data->DecrRefCount();
-    }
-  }
+  LabelLookAheadRelabeler(I **impl);
 
   // Relabels arbitrary Fst. Class L should be a label-lookahead Fst.
   template <class L>
@@ -660,16 +647,60 @@ class LabelLookAheadRelabeler {
 
   // Returns relabeling pairs (cf. relabel.h::Relabel()).
   // Class L should be a label-lookahead Fst.
+  // If 'avoid_collisions' is true, extra pairs are added to
+  // ensure no collisions when relabeling automata that have
+  // labels unseen here.
   template <class L>
-  static void RelabelPairs(const L &mfst, vector<pair<Label, Label> > *pairs) {
+  static void RelabelPairs(const L &mfst, vector<pair<Label, Label> > *pairs,
+                           bool avoid_collisions = false) {
     typename L::Impl *impl = mfst.GetImpl();
     D *data = impl->GetAddOn();
     LabelReachable<A> reachable(data->First() ?
                                   data->First() : data->Second());
-    reachable.RelabelPairs(pairs);
+    reachable.RelabelPairs(pairs, avoid_collisions);
   }
 };
 
+template <class A>
+template <typename I> inline
+LabelLookAheadRelabeler<A>::LabelLookAheadRelabeler(I **impl) {
+  Fst<A> &fst = (*impl)->GetFst();
+  D *data = (*impl)->GetAddOn();
+  const string name = (*impl)->Type();
+  bool is_mutable = fst.Properties(kMutable, false);
+  MutableFst<A> *mfst = 0;
+  if (is_mutable) {
+    mfst = static_cast<MutableFst<A> *>(&fst);
+  } else {
+    mfst = new VectorFst<A>(fst);
+    data->IncrRefCount();
+    delete *impl;
+  }
+  if (data->First()) {  // reach_input
+    LabelReachable<A> reachable(data->First());
+    reachable.Relabel(mfst, true);
+    if (!FLAGS_save_relabel_ipairs.empty()) {
+      vector<pair<Label, Label> > pairs;
+      reachable.RelabelPairs(&pairs, true);
+      WriteLabelPairs(FLAGS_save_relabel_ipairs, pairs);
+    }
+  } else {
+    LabelReachable<A> reachable(data->Second());
+    reachable.Relabel(mfst, false);
+    if (!FLAGS_save_relabel_opairs.empty()) {
+      vector<pair<Label, Label> > pairs;
+      reachable.RelabelPairs(&pairs, true);
+      WriteLabelPairs(FLAGS_save_relabel_opairs, pairs);
+    }
+  }
+  if (!is_mutable) {
+    *impl = new I(*mfst, name);
+    (*impl)->SetAddOn(data);
+    delete mfst;
+    data->DecrRefCount();
+  }
+}
+
 
 // Generic lookahead matcher, templated on the FST definition
 // - a wrapper around pointer to specific one.
@@ -709,45 +740,65 @@ class LookAheadMatcher {
   const Arc& Value() const { return base_->Value(); }
   void Next() { base_->Next(); }
   const F &GetFst() const { return static_cast<const F &>(base_->GetFst()); }
+
   uint64 Properties(uint64 props) const { return base_->Properties(props); }
+
   uint32 Flags() const { return base_->Flags(); }
 
   // Look-ahead methods
   bool LookAheadLabel(Label label) const {
-    LookAheadCheck();
-    LBase *lbase = static_cast<LBase *>(base_);
-    return lbase->LookAheadLabel(label);
+    if (LookAheadCheck()) {
+      LBase *lbase = static_cast<LBase *>(base_);
+      return lbase->LookAheadLabel(label);
+    } else {
+      return true;
+    }
   }
 
   bool LookAheadFst(const Fst<Arc> &fst, StateId s) {
-    LBase *lbase = static_cast<LBase *>(base_);
-    return lbase->LookAheadFst(fst, s);
+    if (LookAheadCheck()) {
+      LBase *lbase = static_cast<LBase *>(base_);
+      return lbase->LookAheadFst(fst, s);
+    } else {
+      return true;
+    }
   }
 
   Weight LookAheadWeight() const {
-    LBase *lbase = static_cast<LBase *>(base_);
-    return lbase->LookAheadWeight();
+    if (LookAheadCheck()) {
+      LBase *lbase = static_cast<LBase *>(base_);
+      return lbase->LookAheadWeight();
+    } else {
+      return Weight::One();
+    }
   }
 
   bool LookAheadPrefix(Arc *arc) const {
-    LBase *lbase = static_cast<LBase *>(base_);
-    return lbase->LookAheadPrefix(arc);
+    if (LookAheadCheck()) {
+      LBase *lbase = static_cast<LBase *>(base_);
+      return lbase->LookAheadPrefix(arc);
+    } else {
+      return false;
+    }
   }
 
   void InitLookAheadFst(const Fst<Arc>& fst, bool copy = false) {
-    LookAheadCheck();
-    LBase *lbase = static_cast<LBase *>(base_);
-    lbase->InitLookAheadFst(fst, copy);
+    if (LookAheadCheck()) {
+      LBase *lbase = static_cast<LBase *>(base_);
+      lbase->InitLookAheadFst(fst, copy);
+    }
   }
 
  private:
-  void LookAheadCheck() const {
+  bool LookAheadCheck() const {
     if (!lookahead_) {
       lookahead_ = base_->Flags() &
           (kInputLookAheadMatcher | kOutputLookAheadMatcher);
-      if (!lookahead_)
-        LOG(FATAL) << "LookAheadMatcher: No look-ahead matcher defined";
+      if (!lookahead_) {
+        FSTERROR() << "LookAheadMatcher: No look-ahead matcher defined";
+      }
     }
+    return lookahead_;
   }
 
   MatcherBase<Arc> *base_;
