@@ -24,7 +24,8 @@
 #include <algorithm>
 #include <set>
 
-#include <fst/fst.h>
+#include <fst/mutable-fst.h>  // for all internal FST accessors
+
 
 namespace fst {
 
@@ -142,7 +143,8 @@ class SortedMatcher : public MatcherBase<typename F::Arc> {
         binary_label_(binary_label),
         match_label_(kNoLabel),
         narcs_(0),
-        loop_(kNoLabel, 0, Weight::One(), kNoStateId) {
+        loop_(kNoLabel, 0, Weight::One(), kNoStateId),
+        error_(false) {
     switch(match_type_) {
       case MATCH_INPUT:
       case MATCH_NONE:
@@ -151,7 +153,9 @@ class SortedMatcher : public MatcherBase<typename F::Arc> {
         swap(loop_.ilabel, loop_.olabel);
         break;
       default:
-        LOG(FATAL) << "SortedMatcher: bad match type";
+        FSTERROR() << "SortedMatcher: bad match type";
+        match_type_ = MATCH_NONE;
+        error_ = true;
     }
   }
 
@@ -163,7 +167,8 @@ class SortedMatcher : public MatcherBase<typename F::Arc> {
         binary_label_(matcher.binary_label_),
         match_label_(kNoLabel),
         narcs_(0),
-        loop_(matcher.loop_) {}
+        loop_(matcher.loop_),
+        error_(matcher.error_) {}
 
   virtual ~SortedMatcher() {
     if (aiter_)
@@ -197,81 +202,39 @@ class SortedMatcher : public MatcherBase<typename F::Arc> {
     if (s_ == s)
       return;
     s_ = s;
-    if (match_type_ == MATCH_NONE)
-        LOG(FATAL) << "SortedMatcher: bad match type";
+    if (match_type_ == MATCH_NONE) {
+      FSTERROR() << "SortedMatcher: bad match type";
+      error_ = true;
+    }
     if (aiter_)
       delete aiter_;
     aiter_ = new ArcIterator<F>(*fst_, s);
     aiter_->SetFlags(kArcNoCache, kArcNoCache);
-    narcs_ = NumArcs(*fst_, s);
+    narcs_ = internal::NumArcs(*fst_, s);
     loop_.nextstate = s;
   }
 
-  bool Find(Label match_label) {
-    current_loop_ = match_label == 0;
-    match_label_ = match_label == kNoLabel ? 0 : match_label;
-    aiter_->SetFlags(
-        match_type_ == MATCH_INPUT ? kArcILabelValue : kArcOLabelValue,
-        kArcValueFlags);
-    if (match_label_ >= binary_label_) {
-      // Binary search for match.
-      size_t low = 0;
-      size_t high = narcs_;
-      while (low < high) {
-        size_t mid = (low + high) / 2;
-        aiter_->Seek(mid);
-        Label label = match_type_ == MATCH_INPUT ?
-            aiter_->Value().ilabel : aiter_->Value().olabel;
-        if (label > match_label_) {
-          high = mid;
-        } else if (label < match_label_) {
-          low = mid + 1;
-        } else {
-          // find first matching label (when non-determinism)
-          for (size_t i = mid; i > low; --i) {
-            aiter_->Seek(i - 1);
-            label = match_type_ == MATCH_INPUT ? aiter_->Value().ilabel :
-                aiter_->Value().olabel;
-            if (label != match_label_) {
-              aiter_->Seek(i);
-              aiter_->SetFlags(kArcValueFlags, kArcValueFlags);
-              return true;
-            }
-          }
-          aiter_->SetFlags(kArcValueFlags, kArcValueFlags);
-          return true;
-        }
-       }
-       return current_loop_;
-    } else {
-      // Linear search for match.
-      for (aiter_->Reset(); !aiter_->Done(); aiter_->Next()) {
-        Label label = match_type_ == MATCH_INPUT ?
-            aiter_->Value().ilabel : aiter_->Value().olabel;
-        if (label == match_label_) {
-          aiter_->SetFlags(kArcValueFlags, kArcValueFlags);
-          return true;
-        }
-        if (label > match_label_)
-          break;
-      }
-      aiter_->SetFlags(kArcValueFlags, kArcValueFlags);
-      return current_loop_;
-    }
-  }
+  bool Find(Label match_label);
 
   bool Done() const {
     if (current_loop_)
       return false;
     if (aiter_->Done())
       return true;
+    aiter_->SetFlags(
+      match_type_ == MATCH_INPUT ? kArcILabelValue : kArcOLabelValue,
+      kArcValueFlags);
     Label label = match_type_ == MATCH_INPUT ?
         aiter_->Value().ilabel : aiter_->Value().olabel;
     return label != match_label_;
   }
 
   const Arc& Value() const {
-    return current_loop_ ? loop_ : aiter_->Value();
+    if (current_loop_) {
+      return loop_;
+    }
+    aiter_->SetFlags(kArcValueFlags, kArcValueFlags);
+    return aiter_->Value();
   }
 
   void Next() {
@@ -283,7 +246,11 @@ class SortedMatcher : public MatcherBase<typename F::Arc> {
 
   virtual const F &GetFst() const { return *fst_; }
 
-  virtual uint64 Properties(uint64 props) const { return props; }
+  virtual uint64 Properties(uint64 inprops) const {
+    uint64 outprops = inprops;
+    if (error_) outprops |= kError;
+    return outprops;
+  }
 
  private:
   virtual void SetState_(StateId s) { SetState(s); }
@@ -301,10 +268,66 @@ class SortedMatcher : public MatcherBase<typename F::Arc> {
   size_t narcs_;                  // Current state arc count
   Arc loop_;                      // For non-consuming symbols
   bool current_loop_;             // Current arc is the implicit loop
+  bool error_;                    // Error encountered
 
   void operator=(const SortedMatcher<F> &);  // Disallow
 };
 
+template <class F> inline
+bool SortedMatcher<F>::Find(Label match_label) {
+  if (error_) {
+    current_loop_ = false;
+    match_label_ = kNoLabel;
+    return false;
+  }
+  current_loop_ = match_label == 0;
+  match_label_ = match_label == kNoLabel ? 0 : match_label;
+  aiter_->SetFlags(
+      match_type_ == MATCH_INPUT ? kArcILabelValue : kArcOLabelValue,
+      kArcValueFlags);
+  if (match_label_ >= binary_label_) {
+    // Binary search for match.
+    size_t low = 0;
+    size_t high = narcs_;
+    while (low < high) {
+      size_t mid = (low + high) / 2;
+      aiter_->Seek(mid);
+      Label label = match_type_ == MATCH_INPUT ?
+          aiter_->Value().ilabel : aiter_->Value().olabel;
+      if (label > match_label_) {
+        high = mid;
+      } else if (label < match_label_) {
+        low = mid + 1;
+      } else {
+        // find first matching label (when non-determinism)
+        for (size_t i = mid; i > low; --i) {
+          aiter_->Seek(i - 1);
+          label = match_type_ == MATCH_INPUT ? aiter_->Value().ilabel :
+              aiter_->Value().olabel;
+          if (label != match_label_) {
+            aiter_->Seek(i);
+            return true;
+          }
+        }
+        return true;
+      }
+    }
+    return current_loop_;
+  } else {
+    // Linear search for match.
+    for (aiter_->Reset(); !aiter_->Done(); aiter_->Next()) {
+      Label label = match_type_ == MATCH_INPUT ?
+          aiter_->Value().ilabel : aiter_->Value().olabel;
+      if (label == match_label_) {
+        return true;
+      }
+      if (label > match_label_)
+        break;
+    }
+    return current_loop_;
+  }
+}
+
 
 // Specifies whether during matching we rewrite both the input and output sides.
 enum MatcherRewriteMode {
@@ -341,11 +364,18 @@ class RhoMatcher : public MatcherBase<typename M::Arc> {
              M *matcher = 0)
       : matcher_(matcher ? matcher : new M(fst, match_type)),
         match_type_(match_type),
-        rho_label_(rho_label) {
-    if (match_type == MATCH_BOTH)
-      LOG(FATAL) << "RhoMatcher: bad match type";
-    if (rho_label == 0)
-      LOG(FATAL) << "RhoMatcher: 0 cannot be used as rho_label";
+        rho_label_(rho_label),
+        error_(false) {
+    if (match_type == MATCH_BOTH) {
+      FSTERROR() << "RhoMatcher: bad match type";
+      match_type_ = MATCH_NONE;
+      error_ = true;
+    }
+    if (rho_label == 0) {
+      FSTERROR() << "RhoMatcher: 0 cannot be used as rho_label";
+      rho_label_ = kNoLabel;
+      error_ = true;
+    }
 
     if (rewrite_mode == MATCHER_REWRITE_AUTO)
       rewrite_both_ = fst.Properties(kAcceptor, true);
@@ -359,7 +389,8 @@ class RhoMatcher : public MatcherBase<typename M::Arc> {
       : matcher_(new M(*matcher.matcher_, safe)),
         match_type_(matcher.match_type_),
         rho_label_(matcher.rho_label_),
-        rewrite_both_(matcher.rewrite_both_) {}
+        rewrite_both_(matcher.rewrite_both_),
+        error_(matcher.error_) {}
 
   virtual ~RhoMatcher() {
     delete matcher_;
@@ -378,7 +409,9 @@ class RhoMatcher : public MatcherBase<typename M::Arc> {
 
   bool Find(Label match_label) {
     if (match_label == rho_label_ && rho_label_ != kNoLabel) {
-      LOG(FATAL) << "RhoMatcher::Find: bad label (rho)";
+      FSTERROR() << "RhoMatcher::Find: bad label (rho)";
+      error_ = true;
+      return false;
     }
     if (matcher_->Find(match_label)) {
       rho_match_ = kNoLabel;
@@ -417,33 +450,7 @@ class RhoMatcher : public MatcherBase<typename M::Arc> {
 
   virtual const FST &GetFst() const { return matcher_->GetFst(); }
 
-  virtual uint64 Properties(uint64 props) const {
-    if (match_type_ == MATCH_NONE) {
-      return props;
-    } else if (match_type_ == MATCH_INPUT) {
-      if (rewrite_both_) {
-        return props & ~(kODeterministic | kNonODeterministic | kString |
-                         kILabelSorted | kNotILabelSorted |
-                         kOLabelSorted | kNotOLabelSorted);
-      } else {
-        return props & ~(kODeterministic | kAcceptor | kString |
-                         kILabelSorted | kNotILabelSorted);
-      }
-    } else if (match_type_ == MATCH_OUTPUT) {
-      if (rewrite_both_) {
-        return props & ~(kIDeterministic | kNonIDeterministic | kString |
-                         kILabelSorted | kNotILabelSorted |
-                         kOLabelSorted | kNotOLabelSorted);
-      } else {
-        return props & ~(kIDeterministic | kAcceptor | kString |
-                         kOLabelSorted | kNotOLabelSorted);
-      }
-    } else {
-      LOG(FATAL) << "RhoMatcher::Properties: Invalid match type: "
-                 << match_type_;
-      return 0;
-    }
-  }
+  virtual uint64 Properties(uint64 props) const;
 
  private:
   virtual void SetState_(StateId s) { SetState(s); }
@@ -459,10 +466,43 @@ class RhoMatcher : public MatcherBase<typename M::Arc> {
   bool has_rho_;          // Are there possibly rhos at the current state?
   Label rho_match_;       // Current label that matches rho transition
   mutable Arc rho_arc_;   // Arc to return when rho match
+  bool error_;            // Error encountered
 
   void operator=(const RhoMatcher<M> &);  // Disallow
 };
 
+template <class M> inline
+uint64 RhoMatcher<M>::Properties(uint64 inprops) const {
+  uint64 outprops = matcher_->Properties(inprops);
+  if (error_) outprops |= kError;
+
+  if (match_type_ == MATCH_NONE) {
+    return outprops;
+  } else if (match_type_ == MATCH_INPUT) {
+    if (rewrite_both_) {
+      return outprops & ~(kODeterministic | kNonODeterministic | kString |
+                       kILabelSorted | kNotILabelSorted |
+                       kOLabelSorted | kNotOLabelSorted);
+    } else {
+      return outprops & ~(kODeterministic | kAcceptor | kString |
+                       kILabelSorted | kNotILabelSorted);
+    }
+  } else if (match_type_ == MATCH_OUTPUT) {
+    if (rewrite_both_) {
+      return outprops & ~(kIDeterministic | kNonIDeterministic | kString |
+                       kILabelSorted | kNotILabelSorted |
+                       kOLabelSorted | kNotOLabelSorted);
+    } else {
+      return outprops & ~(kIDeterministic | kAcceptor | kString |
+                       kOLabelSorted | kNotOLabelSorted);
+    }
+  } else {
+    // Shouldn't ever get here.
+    FSTERROR() << "RhoMatcher:: bad match type: " << match_type_;
+    return 0;
+  }
+}
+
 
 // For any requested label, this matcher considers all transitions
 // that match the label 'sigma_label' (sigma = "any"), and this in
@@ -492,11 +532,18 @@ class SigmaMatcher : public MatcherBase<typename M::Arc> {
                M *matcher = 0)
       : matcher_(matcher ? matcher : new M(fst, match_type)),
         match_type_(match_type),
-        sigma_label_(sigma_label) {
-    if (match_type == MATCH_BOTH)
-      LOG(FATAL) << "SigmaMatcher: bad match type";
-    if (sigma_label == 0)
-      LOG(FATAL) << "SigmaMatcher: 0 cannot be used as sigma_label";
+        sigma_label_(sigma_label),
+        error_(false) {
+    if (match_type == MATCH_BOTH) {
+      FSTERROR() << "SigmaMatcher: bad match type";
+      match_type_ = MATCH_NONE;
+      error_ = true;
+    }
+    if (sigma_label == 0) {
+      FSTERROR() << "SigmaMatcher: 0 cannot be used as sigma_label";
+      sigma_label_ = kNoLabel;
+      error_ = true;
+    }
 
     if (rewrite_mode == MATCHER_REWRITE_AUTO)
       rewrite_both_ = fst.Properties(kAcceptor, true);
@@ -510,7 +557,8 @@ class SigmaMatcher : public MatcherBase<typename M::Arc> {
       : matcher_(new M(*matcher.matcher_, safe)),
         match_type_(matcher.match_type_),
         sigma_label_(matcher.sigma_label_),
-        rewrite_both_(matcher.rewrite_both_) {}
+        rewrite_both_(matcher.rewrite_both_),
+        error_(matcher.error_) {}
 
   virtual ~SigmaMatcher() {
     delete matcher_;
@@ -531,7 +579,9 @@ class SigmaMatcher : public MatcherBase<typename M::Arc> {
   bool Find(Label match_label) {
     match_label_ = match_label;
     if (match_label == sigma_label_ && sigma_label_ != kNoLabel) {
-      LOG(FATAL) << "SigmaMatcher::Find: bad label (sigma)";
+      FSTERROR() << "SigmaMatcher::Find: bad label (sigma)";
+      error_ = true;
+      return false;
     }
     if (matcher_->Find(match_label)) {
       sigma_match_ = kNoLabel;
@@ -579,31 +629,7 @@ class SigmaMatcher : public MatcherBase<typename M::Arc> {
 
   virtual const FST &GetFst() const { return matcher_->GetFst(); }
 
-  virtual uint64 Properties(uint64 props) const {
-    if (match_type_ == MATCH_NONE) {
-      return props;
-    } else if (rewrite_both_) {
-      return props & ~(kIDeterministic | kNonIDeterministic |
-                       kODeterministic | kNonODeterministic |
-                       kILabelSorted | kNotILabelSorted |
-                       kOLabelSorted | kNotOLabelSorted |
-                       kString);
-    } else if (match_type_ == MATCH_INPUT) {
-      return props & ~(kIDeterministic | kNonIDeterministic |
-                       kODeterministic | kNonODeterministic |
-                       kILabelSorted | kNotILabelSorted |
-                       kString | kAcceptor);
-    } else if (match_type_ == MATCH_OUTPUT) {
-      return props & ~(kIDeterministic | kNonIDeterministic |
-                       kODeterministic | kNonODeterministic |
-                       kOLabelSorted | kNotOLabelSorted |
-                       kString | kAcceptor);
-    } else {
-      LOG(FATAL) << "SigmaMatcher::Properties: Invalid match type: "
-                 << match_type_;
-      return 0;
-    }
-  }
+  virtual uint64 Properties(uint64 props) const;
 
 private:
   virtual void SetState_(StateId s) { SetState(s); }
@@ -620,14 +646,45 @@ private:
   Label sigma_match_;      // Current label that matches sigma transition
   mutable Arc sigma_arc_;  // Arc to return when sigma match
   Label match_label_;      // Label being matched
+  bool error_;             // Error encountered
 
   void operator=(const SigmaMatcher<M> &);  // disallow
 };
 
+template <class M> inline
+uint64 SigmaMatcher<M>::Properties(uint64 inprops) const {
+  uint64 outprops = matcher_->Properties(inprops);
+  if (error_) outprops |= kError;
+
+  if (match_type_ == MATCH_NONE) {
+    return outprops;
+  } else if (rewrite_both_) {
+    return outprops & ~(kIDeterministic | kNonIDeterministic |
+                     kODeterministic | kNonODeterministic |
+                     kILabelSorted | kNotILabelSorted |
+                     kOLabelSorted | kNotOLabelSorted |
+                     kString);
+  } else if (match_type_ == MATCH_INPUT) {
+    return outprops & ~(kIDeterministic | kNonIDeterministic |
+                     kODeterministic | kNonODeterministic |
+                     kILabelSorted | kNotILabelSorted |
+                     kString | kAcceptor);
+  } else if (match_type_ == MATCH_OUTPUT) {
+    return outprops & ~(kIDeterministic | kNonIDeterministic |
+                     kODeterministic | kNonODeterministic |
+                     kOLabelSorted | kNotOLabelSorted |
+                     kString | kAcceptor);
+  } else {
+    // Shouldn't ever get here.
+    FSTERROR() << "SigmaMatcher:: bad match type: " << match_type_;
+    return 0;
+  }
+}
+
 
 // For any requested label that doesn't match at a state, this matcher
-// considers the *unique* transition that match the label 'phi_label'
-// (phi = 'fail'), and recursively look for a match at its
+// considers the *unique* transition that matches the label 'phi_label'
+// (phi = 'fail'), and recursively looks for a match at its
 // destination.  When 'phi_loop' is true, if no match is found but a
 // phi self-loop is found, then the phi transition found is returned
 // with the phi_label rewritten as the requested label (both sides if
@@ -658,11 +715,13 @@ class PhiMatcher : public MatcherBase<typename M::Arc> {
         match_type_(match_type),
         phi_label_(phi_label),
         state_(kNoStateId),
-        phi_loop_(phi_loop) {
-    if (match_type == MATCH_BOTH)
-      LOG(FATAL) << "PhiMatcher: bad match type";
-    if (phi_label == 0)
-      LOG(FATAL) << "PhiMatcher: 0 cannot be used as phi_label";
+        phi_loop_(phi_loop),
+        error_(false) {
+    if (match_type == MATCH_BOTH) {
+      FSTERROR() << "PhiMatcher: bad match type";
+      match_type_ = MATCH_NONE;
+      error_ = true;
+    }
 
     if (rewrite_mode == MATCHER_REWRITE_AUTO)
       rewrite_both_ = fst.Properties(kAcceptor, true);
@@ -678,7 +737,8 @@ class PhiMatcher : public MatcherBase<typename M::Arc> {
         phi_label_(matcher.phi_label_),
         rewrite_both_(matcher.rewrite_both_),
         state_(kNoStateId),
-        phi_loop_(matcher.phi_loop_) {}
+        phi_loop_(matcher.phi_loop_),
+        error_(matcher.error_) {}
 
   virtual ~PhiMatcher() {
     delete matcher_;
@@ -696,42 +756,22 @@ class PhiMatcher : public MatcherBase<typename M::Arc> {
     has_phi_ = phi_label_ != kNoLabel;
   }
 
-  bool Find(Label match_label) {
-    if (match_label == phi_label_ && phi_label_ != kNoLabel) {
-      LOG(FATAL) << "PhiMatcher::Find: bad label (phi)";
-    }
-    matcher_->SetState(state_);
-    phi_match_ = kNoLabel;
-    phi_weight_ = Weight::One();
-    if (!has_phi_ || match_label == 0 || match_label == kNoLabel)
-      return matcher_->Find(match_label);
-    StateId state = state_;
-    while (!matcher_->Find(match_label)) {
-      if (!matcher_->Find(phi_label_))
-        return false;
-      if (phi_loop_ && matcher_->Value().nextstate == state) {
-        phi_match_ = match_label;
-        return true;
-      }
-      phi_weight_ = Times(phi_weight_, matcher_->Value().weight);
-      state = matcher_->Value().nextstate;
-      matcher_->Next();
-      if (!matcher_->Done())
-        LOG(FATAL) << "PhiMatcher: phi non-determinism not supported";
-      matcher_->SetState(state);
-    }
-    return true;
-  }
+  bool Find(Label match_label);
 
   bool Done() const { return matcher_->Done(); }
 
   const Arc& Value() const {
     if ((phi_match_ == kNoLabel) && (phi_weight_ == Weight::One())) {
       return matcher_->Value();
+    } else if (phi_match_ == 0) {  // Virtual epsilon loop
+      phi_arc_ = Arc(kNoLabel, 0, Weight::One(), state_);
+      if (match_type_ == MATCH_OUTPUT)
+        swap(phi_arc_.ilabel, phi_arc_.olabel);
+      return phi_arc_;
     } else {
       phi_arc_ = matcher_->Value();
       phi_arc_.weight = Times(phi_weight_, phi_arc_.weight);
-      if (phi_match_ != kNoLabel) {
+      if (phi_match_ != kNoLabel) {  // Phi loop match
         if (rewrite_both_) {
           if (phi_arc_.ilabel == phi_label_)
             phi_arc_.ilabel = phi_match_;
@@ -751,35 +791,7 @@ class PhiMatcher : public MatcherBase<typename M::Arc> {
 
   virtual const FST &GetFst() const { return matcher_->GetFst(); }
 
-  virtual uint64 Properties(uint64 props) const {
-    if (match_type_ == MATCH_NONE) {
-      return props;
-    } else if (match_type_ == MATCH_INPUT) {
-      if (rewrite_both_) {
-        return props & ~(kODeterministic | kNonODeterministic | kString |
-                         kILabelSorted | kNotILabelSorted |
-                         kOLabelSorted | kNotOLabelSorted);
-      } else {
-        return props & ~(kODeterministic | kAcceptor | kString |
-                         kILabelSorted | kNotILabelSorted |
-                         kOLabelSorted | kNotOLabelSorted);
-      }
-    } else if (match_type_ == MATCH_OUTPUT) {
-      if (rewrite_both_) {
-        return props & ~(kIDeterministic | kNonIDeterministic | kString |
-                         kILabelSorted | kNotILabelSorted |
-                         kOLabelSorted | kNotOLabelSorted);
-      } else {
-        return props & ~(kIDeterministic | kAcceptor | kString |
-                         kILabelSorted | kNotILabelSorted |
-                         kOLabelSorted | kNotOLabelSorted);
-      }
-    } else {
-      LOG(FATAL) << "PhiMatcher::Properties: Invalid match type: "
-                 << match_type_;
-      return 0;
-    }
-  }
+  virtual uint64 Properties(uint64 props) const;
 
 private:
   virtual void SetState_(StateId s) { SetState(s); }
@@ -799,10 +811,99 @@ private:
   Weight phi_weight_;     // Product of the weights of phi transitions taken
   bool phi_loop_;         // When true, phi self-loop are allowed and treated
                           // as rho (required for Aho-Corasick)
+  bool error_;             // Error encountered
 
   void operator=(const PhiMatcher<M> &);  // disallow
 };
 
+template <class M> inline
+bool PhiMatcher<M>::Find(Label match_label) {
+  if (match_label == phi_label_ && phi_label_ != kNoLabel && phi_label_ != 0) {
+    FSTERROR() << "PhiMatcher::Find: bad label (phi): " << phi_label_;
+    error_ = true;
+    return false;
+  }
+  matcher_->SetState(state_);
+  phi_match_ = kNoLabel;
+  phi_weight_ = Weight::One();
+  if (phi_label_ == 0) {          // When 'phi_label_ == 0',
+    if (match_label == kNoLabel)  // there are no more true epsilon arcs,
+      return false;
+    if (match_label == 0) {       // but virtual eps loop need to be returned
+      if (!matcher_->Find(kNoLabel)) {
+        return matcher_->Find(0);
+      } else {
+        phi_match_ = 0;
+        return true;
+      }
+    }
+  }
+  if (!has_phi_ || match_label == 0 || match_label == kNoLabel)
+    return matcher_->Find(match_label);
+  StateId state = state_;
+  while (!matcher_->Find(match_label)) {
+    // Look for phi transition (if phi_label_ == 0, we need to look
+    // for -1 to avoid getting the virtual self-loop)
+    if (!matcher_->Find(phi_label_ == 0 ? -1 : phi_label_))
+      return false;
+    if (phi_loop_ && matcher_->Value().nextstate == state) {
+      phi_match_ = match_label;
+      return true;
+    }
+    phi_weight_ = Times(phi_weight_, matcher_->Value().weight);
+    state = matcher_->Value().nextstate;
+    matcher_->Next();
+    if (!matcher_->Done()) {
+      FSTERROR() << "PhiMatcher: phi non-determinism not supported";
+      error_ = true;
+    }
+    matcher_->SetState(state);
+  }
+  return true;
+}
+
+template <class M> inline
+uint64 PhiMatcher<M>::Properties(uint64 inprops) const {
+  uint64 outprops = matcher_->Properties(inprops);
+  if (error_) outprops |= kError;
+
+  if (match_type_ == MATCH_NONE) {
+    return outprops;
+  } else if (match_type_ == MATCH_INPUT) {
+    if (phi_label_ == 0) {
+      outprops &= ~kEpsilons | ~kIEpsilons | ~kOEpsilons;
+      outprops |= kNoEpsilons | kNoIEpsilons;
+    }
+    if (rewrite_both_) {
+      return outprops & ~(kODeterministic | kNonODeterministic | kString |
+                       kILabelSorted | kNotILabelSorted |
+                       kOLabelSorted | kNotOLabelSorted);
+    } else {
+      return outprops & ~(kODeterministic | kAcceptor | kString |
+                       kILabelSorted | kNotILabelSorted |
+                       kOLabelSorted | kNotOLabelSorted);
+    }
+  } else if (match_type_ == MATCH_OUTPUT) {
+    if (phi_label_ == 0) {
+      outprops &= ~kEpsilons | ~kIEpsilons | ~kOEpsilons;
+      outprops |= kNoEpsilons | kNoOEpsilons;
+    }
+    if (rewrite_both_) {
+      return outprops & ~(kIDeterministic | kNonIDeterministic | kString |
+                       kILabelSorted | kNotILabelSorted |
+                       kOLabelSorted | kNotOLabelSorted);
+    } else {
+      return outprops & ~(kIDeterministic | kAcceptor | kString |
+                       kILabelSorted | kNotILabelSorted |
+                       kOLabelSorted | kNotOLabelSorted);
+    }
+  } else {
+    // Shouldn't ever get here.
+    FSTERROR() << "PhiMatcher:: bad match type: " << match_type_;
+    return 0;
+  }
+}
+
 
 //
 // MULTI-EPS MATCHER FLAGS
@@ -820,9 +921,8 @@ const uint32 kMultiEpsLoop =  0x00000002;
 // default, the underlying matcher is constructed by
 // MultiEpsMatcher. The user can instead pass in this object; in that
 // case, MultiEpsMatcher takes its ownership iff 'own_matcher' is
-// true. Second template argument determines internal STL storage
-// class for multi-epsilons.
-template <class M, class S = set<typename M::Arc::Label> >
+// true.
+template <class M>
 class MultiEpsMatcher {
  public:
   typedef typename M::FST FST;
@@ -836,9 +936,7 @@ class MultiEpsMatcher {
                   M *matcher = 0, bool own_matcher = true)
       : matcher_(matcher ? matcher : new M(fst, match_type)),
         flags_(flags),
-        own_matcher_(matcher ?  own_matcher : true),
-        min_multi_eps_label_(kNoLabel),
-        max_multi_eps_label_(kNoLabel) {
+        own_matcher_(matcher ?  own_matcher : true) {
     if (match_type == MATCH_INPUT) {
       loop_.ilabel = kNoLabel;
       loop_.olabel = 0;
@@ -855,9 +953,7 @@ class MultiEpsMatcher {
         flags_(matcher.flags_),
         own_matcher_(true),
         multi_eps_labels_(matcher.multi_eps_labels_),
-        loop_(matcher.loop_),
-        min_multi_eps_label_(matcher.min_multi_eps_label_),
-        max_multi_eps_label_(matcher.max_multi_eps_label_) {
+        loop_(matcher.loop_) {
     loop_.nextstate = kNoStateId;
   }
 
@@ -877,38 +973,7 @@ class MultiEpsMatcher {
     loop_.nextstate = s;
   }
 
-  bool Find(Label match_label) {
-    multi_eps_iter_ = multi_eps_labels_.end();
-    current_loop_ = false;
-    bool ret;
-    if (match_label == 0) {
-      ret = matcher_->Find(0);
-    } else if (match_label == kNoLabel) {
-      if (flags_ & kMultiEpsList) {
-        // return all non-consuming arcs (incl. epsilon)
-        multi_eps_iter_ = multi_eps_labels_.begin();
-        while ((multi_eps_iter_ != multi_eps_labels_.end()) &&
-               !matcher_->Find(*multi_eps_iter_))
-          ++multi_eps_iter_;
-        if (multi_eps_iter_ != multi_eps_labels_.end())
-          ret = true;
-        else
-          ret = matcher_->Find(kNoLabel);
-      } else {
-        // return all epsilon arcs
-        ret = matcher_->Find(kNoLabel);
-      }
-    } else if ((flags_ & kMultiEpsLoop) &&
-               IsMultiEps(multi_eps_labels_, match_label)) {
-      // return 'implicit' loop
-      current_loop_ = true;
-      ret = true;
-    } else {
-      ret = matcher_->Find(match_label);
-    }
-    done_ = !ret;
-    return ret;
-  }
+  bool Find(Label match_label);
 
   bool Done() const {
     return done_;
@@ -922,12 +987,12 @@ class MultiEpsMatcher {
     if (!current_loop_) {
       matcher_->Next();
       done_ = matcher_->Done();
-      if (done_ && multi_eps_iter_ != multi_eps_labels_.end()) {
+      if (done_ && multi_eps_iter_ != multi_eps_labels_.End()) {
         ++multi_eps_iter_;
-        while ((multi_eps_iter_ != multi_eps_labels_.end()) &&
+        while ((multi_eps_iter_ != multi_eps_labels_.End()) &&
                !matcher_->Find(*multi_eps_iter_))
           ++multi_eps_iter_;
-        if (multi_eps_iter_ != multi_eps_labels_.end())
+        if (multi_eps_iter_ != multi_eps_labels_.End())
           done_ = false;
         else
           done_ = !matcher_->Find(kNoLabel);
@@ -940,58 +1005,78 @@ class MultiEpsMatcher {
 
   const FST &GetFst() const { return matcher_->GetFst(); }
 
-  uint64 Properties(uint64 props) const { return props; }
+  uint64 Properties(uint64 props) const { return matcher_->Properties(props); }
 
   uint32 Flags() const { return matcher_->Flags(); }
 
   void AddMultiEpsLabel(Label label) {
-    if (label == 0)
-      LOG(FATAL) << "MultiEpsMatcher: Bad multi-eps label: 0";
-    multi_eps_labels_.insert(multi_eps_labels_.end(), label);
-    if (min_multi_eps_label_ == kNoLabel || label < min_multi_eps_label_)
-      min_multi_eps_label_ = label;
-    if (max_multi_eps_label_ == kNoLabel || label > max_multi_eps_label_)
-      max_multi_eps_label_ = label;
+    if (label == 0) {
+      FSTERROR() << "MultiEpsMatcher: Bad multi-eps label: 0";
+    } else {
+      multi_eps_labels_.Insert(label);
+    }
   }
 
   void ClearMultiEpsLabels() {
-    multi_eps_labels_.clear();
-    min_multi_eps_label_ = max_multi_eps_label_ = kNoLabel;
+    multi_eps_labels_.Clear();
   }
 
 private:
-  // Generic case - linear lookup
-  template <class T>
-  bool IsMultiEps(const T &multi_eps_labels, Label label) const {
-    if (min_multi_eps_label_ == kNoLabel ||
-        label < min_multi_eps_label_ || label > max_multi_eps_label_)
-      return false;
-    return find(multi_eps_labels.begin(), multi_eps_labels.end(), label)
-        != multi_eps_labels.end();
-  }
-
   // Specialized for 'set' - log lookup
   bool IsMultiEps(const set<Label> &multi_eps_labels, Label label) const {
-    if (min_multi_eps_label_ == kNoLabel ||
-        label < min_multi_eps_label_ || label > max_multi_eps_label_)
-      return false;
-    return multi_eps_labels.find(label) != multi_eps_labels.end();
+    return multi_eps_labels.Find(label) != multi_eps_labels.end();
   }
 
   M *matcher_;
   uint32 flags_;
   bool own_matcher_;             // Does this class delete the matcher?
-  S multi_eps_labels_;           // Multi-eps label set
-  typename S::const_iterator multi_eps_iter_;
+
+  // Multi-eps label set
+  CompactSet<Label, kNoLabel> multi_eps_labels_;
+  typename CompactSet<Label, kNoLabel>::const_iterator multi_eps_iter_;
+
   bool current_loop_;            // Current arc is the implicit loop
   mutable Arc loop_;             // For non-consuming symbols
   bool done_;                    // Matching done
-  Label min_multi_eps_label_;    // For faster IsMultiEps check
-  Label max_multi_eps_label_;    // For faster IsMultiEps check
 
   void operator=(const MultiEpsMatcher<M> &);  // Disallow
 };
 
+template <class M> inline
+bool MultiEpsMatcher<M>::Find(Label match_label) {
+  multi_eps_iter_ = multi_eps_labels_.End();
+  current_loop_ = false;
+  bool ret;
+  if (match_label == 0) {
+    ret = matcher_->Find(0);
+  } else if (match_label == kNoLabel) {
+    if (flags_ & kMultiEpsList) {
+      // return all non-consuming arcs (incl. epsilon)
+      multi_eps_iter_ = multi_eps_labels_.Begin();
+      while ((multi_eps_iter_ != multi_eps_labels_.End()) &&
+             !matcher_->Find(*multi_eps_iter_))
+        ++multi_eps_iter_;
+      if (multi_eps_iter_ != multi_eps_labels_.End())
+        ret = true;
+      else
+        ret = matcher_->Find(kNoLabel);
+    } else {
+      // return all epsilon arcs
+      ret = matcher_->Find(kNoLabel);
+    }
+  } else if ((flags_ & kMultiEpsLoop) &&
+             multi_eps_labels_.Find(match_label) != multi_eps_labels_.End()) {
+    // return 'implicit' loop
+    current_loop_ = true;
+    ret = true;
+  } else {
+    ret = matcher_->Find(match_label);
+  }
+  done_ = !ret;
+  return ret;
+}
+
+
 // Generic matcher, templated on the FST definition
 // - a wrapper around pointer to specific one.
 // Here is a typical use: \code
