@@ -23,6 +23,7 @@
 
 #include <tr1/unordered_map>
 using std::tr1::unordered_map;
+using std::tr1::unordered_multimap;
 #include <fst/slist.h>
 #include <stack>
 #include <string>
@@ -36,12 +37,12 @@ using std::vector;
 #include <fst/connect.h>
 #include <fst/factor-weight.h>
 #include <fst/invert.h>
-#include <fst/map.h>
 #include <fst/prune.h>
 #include <fst/queue.h>
 #include <fst/shortest-distance.h>
 #include <fst/topsort.h>
 
+
 namespace fst {
 
 template <class Arc, class Queue>
@@ -88,6 +89,9 @@ class RmEpsilonState {
   // Returns final weight of expanded state.
   const Weight &Final() const { return final_; }
 
+  // Return true if an error has occured.
+  bool Error() const { return sd_state_.Error(); }
+
  private:
   static const size_t kPrime0 = 7853;
   static const size_t kPrime1 = 7867;
@@ -106,7 +110,6 @@ class RmEpsilonState {
   class ElementKey {
    public:
     size_t operator()(const Element& e) const {
-      return static_cast<size_t>(e.nextstate);
       return static_cast<size_t>(e.nextstate +
                                  e.ilabel * kPrime0 +
                                  e.olabel * kPrime1);
@@ -155,10 +158,12 @@ const size_t RmEpsilonState<Arc, Queue>::kPrime1;
 
 template <class Arc, class Queue>
 void RmEpsilonState<Arc,Queue>::Expand(typename Arc::StateId source) {
-   sd_state_.ShortestDistance(source);
-   eps_queue_.push(source);
    final_ = Weight::Zero();
    arcs_.clear();
+   sd_state_.ShortestDistance(source);
+   if (sd_state_.Error())
+     return;
+   eps_queue_.push(source);
 
    while (!eps_queue_.empty()) {
      StateId state = eps_queue_.top();
@@ -227,9 +232,27 @@ void RmEpsilon(MutableFst<Arc> *fst,
   typedef typename Arc::Weight Weight;
   typedef typename Arc::Label Label;
 
+  if (fst->Start() == kNoStateId) {
+    return;
+  }
+
+  // 'noneps_in[s]' will be set to true iff 's' admits a non-epsilon
+  // incoming transition or is the start state.
+  vector<bool> noneps_in(fst->NumStates(), false);
+  noneps_in[fst->Start()] = true;
+  for (StateId i = 0; i < fst->NumStates(); ++i) {
+    for (ArcIterator<Fst<Arc> > aiter(*fst, i);
+         !aiter.Done();
+         aiter.Next()) {
+      if (aiter.Value().ilabel != 0 || aiter.Value().olabel != 0)
+        noneps_in[aiter.Value().nextstate] = true;
+    }
+  }
+
   // States sorted in topological order when (acyclic) or generic
   // topological order (cyclic).
   vector<StateId> states;
+  states.reserve(fst->NumStates());
 
   if (fst->Properties(kTopSorted, false) & kTopSorted) {
     for (StateId i = 0; i < fst->NumStates(); i++)
@@ -239,8 +262,12 @@ void RmEpsilon(MutableFst<Arc> *fst,
     bool acyclic;
     TopOrderVisitor<Arc> top_order_visitor(&order, &acyclic);
     DfsVisit(*fst, &top_order_visitor, EpsilonArcFilter<Arc>());
-    if (!acyclic)
-      LOG(FATAL) << "RmEpsilon: not acyclic though property bit is set";
+    // Sanity check: should be acyclic if property bit is set.
+    if(!acyclic) {
+      FSTERROR() << "RmEpsilon: inconsistent acyclic property bit";
+      fst->SetProperties(kError, kError);
+      return;
+    }
     states.resize(order.size());
     for (StateId i = 0; i < order.size(); i++)
       states[order[i]] = i;
@@ -267,19 +294,29 @@ void RmEpsilon(MutableFst<Arc> *fst,
   while (!states.empty()) {
     StateId state = states.back();
     states.pop_back();
+    if (!noneps_in[state])
+      continue;
     rmeps_state.Expand(state);
     fst->SetFinal(state, rmeps_state.Final());
     fst->DeleteArcs(state);
     vector<Arc> &arcs = rmeps_state.Arcs();
+    fst->ReserveArcs(state, arcs.size());
     while (!arcs.empty()) {
       fst->AddArc(state, arcs.back());
       arcs.pop_back();
     }
   }
 
-  fst->SetProperties(RmEpsilonProperties(
-                         fst->Properties(kFstProperties, false)),
-                     kFstProperties);
+  for (StateId s = 0; s < fst->NumStates(); ++s) {
+    if (!noneps_in[s])
+      fst->DeleteArcs(s);
+  }
+
+  if(rmeps_state.Error())
+    fst->SetProperties(kError, kError);
+  fst->SetProperties(
+      RmEpsilonProperties(fst->Properties(kFstProperties, false)),
+      kFstProperties);
 
   if (opts.weight_threshold != Weight::Zero() ||
       opts.state_threshold != kNoStateId)
@@ -289,30 +326,6 @@ void RmEpsilon(MutableFst<Arc> *fst,
     Connect(fst);
 }
 
-// Similar to the above, but can optionally perform the operation
-// in the reverse direction.
-template <class Arc, class Queue>
-void RmEpsilon(const Fst<Arc> &ifst, MutableFst<Arc> *ofst,
-               vector<typename Arc::Weight> *distance,
-               bool reverse,
-               const RmEpsilonOptions<Arc, Queue> &opts) {
-  if (reverse) {
-    VectorFst< ReverseArc<Arc> > rfst;
-    Reverse(ifst, &rfst);
-
-    RmEpsilonOptions<ReverseArc<Arc>, Queue> reverse_opts(
-        opts.state_queue, opts.delta, opts.connect, opts.weight_threshold,
-        opts.source);
-
-    RmEpsilon(&rfst, distance, reverse_opts);
-    Reverse(rfst, ofst);
-    RmEpsilon(ofst, distance, opts);
-  } else {
-    *ofst = ifst;
-    RmEpsilon(ofst, distance, opts);
-  }
-}
-
 // Removes epsilon-transitions (when both the input and output label
 // are an epsilon) from a transducer. The result will be an equivalent
 // FST that has no such epsilon transitions. This version modifies its
@@ -367,11 +380,10 @@ class RmEpsilonFstImpl : public CacheImpl<A> {
  public:
   using FstImpl<A>::SetType;
   using FstImpl<A>::SetProperties;
-  using FstImpl<A>::Properties;
   using FstImpl<A>::SetInputSymbols;
   using FstImpl<A>::SetOutputSymbols;
 
-  using CacheBaseImpl< CacheState<A> >::AddArc;
+  using CacheBaseImpl< CacheState<A> >::PushArc;
   using CacheBaseImpl< CacheState<A> >::HasArcs;
   using CacheBaseImpl< CacheState<A> >::HasFinal;
   using CacheBaseImpl< CacheState<A> >::HasStart;
@@ -449,12 +461,15 @@ class RmEpsilonFstImpl : public CacheImpl<A> {
     return CacheImpl<A>::NumOutputEpsilons(s);
   }
 
-  /*
-  void InitStateIterator(StateIteratorData<A> *data) {
-    LOG(FATAL) << "RmEpsilonFst: InitStateIterator defined in the Fst "
-               << "interface not its implementation.";
+  uint64 Properties() const { return Properties(kFstProperties); }
+
+  // Set error if found; return FST impl properties.
+  uint64 Properties(uint64 mask) const {
+    if ((mask & kError) &&
+        (fst_->Properties(kError, false) || rmeps_state_.Error()))
+      SetProperties(kError, kError);
+    return FstImpl<A>::Properties(mask);
   }
-  */
 
   void InitArcIterator(StateId s, ArcIteratorData<A> *data) {
     if (!HasArcs(s))
@@ -467,7 +482,7 @@ class RmEpsilonFstImpl : public CacheImpl<A> {
     SetFinal(s, rmeps_state_.Final());
     vector<A> &arcs = rmeps_state_.Arcs();
     while (!arcs.empty()) {
-      AddArc(s, arcs.back());
+      PushArc(s, arcs.back());
       arcs.pop_back();
     }
     SetArcs(s);
