@@ -22,9 +22,11 @@
 #ifndef FST_LIB_VISIT_H__
 #define FST_LIB_VISIT_H__
 
+
 #include <fst/arcfilter.h>
 #include <fst/mutable-fst.h>
 
+
 namespace fst {
 
 // Visitor Interface - class determines actions taken during a visit.
@@ -164,7 +166,8 @@ void Visit(const Fst<Arc> &fst, V *visitor, Q *queue, ArcFilter filter) {
     // Finds next tree root
     for (root = root == start ? 0 : root + 1;
          root < nstates && state_status[root] != kWhiteState;
-         ++root);
+         ++root) {
+    }
 
     // Check for a state beyond the largest known state
     if (!expanded && root == nstates) {
