@@ -20,6 +20,7 @@
 // Classes to provide symbol-to-integer and integer-to-symbol mappings.
 
 #include <fst/symbol-table.h>
+
 #include <fst/util.h>
 
 DEFINE_bool(fst_compat_symbols, true,
@@ -50,8 +51,10 @@ SymbolTableImpl* SymbolTableImpl::ReadText(istream &strm,
     if (col.size() == 0)  // empty line
       continue;
     if (col.size() != 2) {
-      LOG(ERROR) << "SymbolTable::ReadText: Bad number of columns (skipping), "
-                 << "file = " << filename << ", line = " << nline;
+      LOG(ERROR) << "SymbolTable::ReadText: Bad number of columns ("
+                 << col.size() << " skipping), "
+                 << "file = " << filename << ", line = " << nline
+                 << ":<" << line << ">";
       continue;
     }
     const char *symbol = col[0];
@@ -72,33 +75,41 @@ SymbolTableImpl* SymbolTableImpl::ReadText(istream &strm,
 }
 
 void SymbolTableImpl::MaybeRecomputeCheckSum() const {
-  if (check_sum_finalized_)
-    return;
+  {
+    ReaderMutexLock check_sum_lock(&check_sum_mutex_);
+    if (check_sum_finalized_)
+      return;
+  }
+
+  // We'll aquire an exclusive lock to recompute the checksums.
+  MutexLock check_sum_lock(&check_sum_mutex_);
+  if (check_sum_finalized_)  // Another thread (coming in around the same time
+    return;                  // might have done it already).  So we recheck.
 
   // Calculate the original label-agnostic check sum.
-  check_sum_.Reset();
+  CheckSummer check_sum;
   for (int64 i = 0; i < symbols_.size(); ++i)
-    check_sum_.Update(symbols_[i], strlen(symbols_[i]) + 1);
-  check_sum_string_ = check_sum_.Digest();
+    check_sum.Update(symbols_[i], strlen(symbols_[i]) + 1);
+  check_sum_string_ = check_sum.Digest();
 
   // Calculate the safer, label-dependent check sum.
-  labeled_check_sum_.Reset();
+  CheckSummer labeled_check_sum;
   for (int64 key = 0; key < dense_key_limit_; ++key) {
-    char line[kLineLen];
-    snprintf(line, kLineLen, "%s\t%lld", symbols_[key], key);
-    labeled_check_sum_.Update(line);
+    ostringstream line;
+    line << symbols_[key] << '\t' << key;
+    labeled_check_sum.Update(line.str().data(), line.str().size());
   }
   for (map<int64, const char*>::const_iterator it =
        key_map_.begin();
        it != key_map_.end();
        ++it) {
     if (it->first >= dense_key_limit_) {
-      char line[kLineLen];
-      snprintf(line, kLineLen, "%s\t%lld", it->second, it->first);
-      labeled_check_sum_.Update(line);
+      ostringstream line;
+      line << it->second << '\t' << it->first;
+      labeled_check_sum.Update(line.str().data(), line.str().size());
     }
   }
-  labeled_check_sum_string_ = labeled_check_sum_.Digest();
+  labeled_check_sum_string_ = labeled_check_sum.Digest();
 
   check_sum_finalized_ = true;
 }
@@ -155,8 +166,11 @@ SymbolTableImpl* SymbolTableImpl::Read(istream &strm,
   ReadType(strm, &impl->available_key_);
   int64 size;
   ReadType(strm, &size);
-  if (!strm)
+  if (!strm) {
     LOG(ERROR) << "SymbolTable::Read: read failed";
+    delete impl;
+    return 0;
+  }
 
   string symbol;
   int64 key;
@@ -164,8 +178,11 @@ SymbolTableImpl* SymbolTableImpl::Read(istream &strm,
   for (size_t i = 0; i < size; ++i) {
     ReadType(strm, &symbol);
     ReadType(strm, &key);
-    if (!strm)
+    if (!strm) {
       LOG(ERROR) << "SymbolTable::Read: read failed";
+      delete impl;
+      return 0;
+    }
 
     char *csymbol = new char[symbol.size() + 1];
     strcpy(csymbol, symbol.c_str());
@@ -190,15 +207,23 @@ bool SymbolTableImpl::Write(ostream &strm) const {
   int64 size = symbols_.size();
   WriteType(strm, size);
   // first write out dense keys
-  for (int64 i = 0; i < dense_key_limit_; ++i) {
+  int64 i = 0;
+  for (; i < dense_key_limit_; ++i) {
     WriteType(strm, string(symbols_[i]));
     WriteType(strm, i);
   }
   // next write out the remaining non densely packed keys
-  for (map<int64, const char*>::const_iterator it =
-           key_map_.begin(); it != key_map_.end(); ++it) {
-    WriteType(strm, string(it->second));
-    WriteType(strm, it->first);
+  for (map<const char *, int64, StrCmp>::const_iterator it =
+           symbol_map_.begin(); it != symbol_map_.end(); ++it) {
+    if ((it->second >= 0) && (it->second < dense_key_limit_))
+      continue;
+    WriteType(strm, string(it->first));
+    WriteType(strm, it->second);
+    ++i;
+  }
+  if (i != size) {
+    LOG(ERROR) << "SymbolTable::Write:  write failed";
+    return false;
   }
   strm.flush();
   if (!strm) {
@@ -218,11 +243,10 @@ void SymbolTable::AddTable(const SymbolTable& table) {
 
 bool SymbolTable::WriteText(ostream &strm) const {
   for (SymbolTableIterator iter(*this); !iter.Done(); iter.Next()) {
-    char line[kLineLen];
-    snprintf(line, kLineLen, "%s%c%lld\n",
-             iter.Symbol().c_str(), FLAGS_fst_field_separator[0],
-             iter.Value());
-    strm.write(line, strlen(line));
+    ostringstream line;
+    line << iter.Symbol() << FLAGS_fst_field_separator[0] << iter.Value()
+         << '\n';
+    strm.write(line.str().data(), line.str().length());
   }
   return true;
 }
