@@ -34,6 +34,7 @@ namespace script {
 
 REGISTER_FST_CLASSES(StdArc);
 REGISTER_FST_CLASSES(LogArc);
+REGISTER_FST_CLASSES(Log64Arc);
 
 //
 //  FST CLASS METHODS
@@ -73,15 +74,44 @@ FstClass *FstClass::Read(const string &fname) {
     ifstream in(fname.c_str(), ifstream::in | ifstream::binary);
     return ReadFst<FstClass>(in, fname);
   } else {
-    return ReadFst<FstClass>(std::cin, "stdin");
+    return ReadFst<FstClass>(cin, "standard input");
+  }
+}
+
+FstClass *FstClass::Read(istream &istr, const string &source) {
+  return ReadFst<FstClass>(istr, source);
+}
+
+//
+//  MUTABLE FST CLASS METHODS
+//
+
+MutableFstClass *MutableFstClass::Read(const string &fname, bool convert) {
+  if (convert == false) {
+    if (!fname.empty()) {
+      ifstream in(fname.c_str(), ifstream::in | ifstream::binary);
+      return ReadFst<MutableFstClass>(in, fname);
+    } else {
+      return ReadFst<MutableFstClass>(cin, "standard input");
+    }
+  } else {  // Converts to VectorFstClass if not mutable.
+    FstClass *ifst = FstClass::Read(fname);
+    if (!ifst) return 0;
+    if (ifst->Properties(fst::kMutable, false)) {
+      return static_cast<MutableFstClass *>(ifst);
+    } else {
+      MutableFstClass *ofst = new VectorFstClass(*ifst);
+      delete ifst;
+      return ofst;
+    }
   }
 }
 
 //
-// VECTOR FST CONSTRUCTORS
+// VECTOR FST CLASS METHODS
 //
 
-const IORegistration<VectorFstClass>::Entry &GetVFSTRegisterEntry(
+IORegistration<VectorFstClass>::Entry GetVFSTRegisterEntry(
     const string &arc_type) {
   IORegistration<VectorFstClass>::Register *reg =
       IORegistration<VectorFstClass>::Register::GetRegister();
@@ -102,5 +132,14 @@ VectorFstClass::VectorFstClass(const FstClass &other)
 VectorFstClass::VectorFstClass(const string &arc_type)
     : MutableFstClass(GetVFSTRegisterEntry(arc_type).creator()) { }
 
+VectorFstClass *VectorFstClass::Read(const string &fname) {
+  if (!fname.empty()) {
+    ifstream in(fname.c_str(), ifstream::in | ifstream::binary);
+    return ReadFst<VectorFstClass>(in, fname);
+  } else {
+    return ReadFst<VectorFstClass>(cin, "standard input");
+  }
+}
+
 }  // namespace script
 }  // namespace fst
